"""Test src/bitwarden_workflow_linter/rules/job_environment_prefix."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.job_environment_prefix import (
    RuleJobEnvironmentPrefix,
)

yaml = YAML()


@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    env:
        _TEST_ENV: "test"
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="no_env_workflow")
def fixture_no_env_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="missing_prefix_workflow")
def fixture_missing_prefix_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    env:
        TEST_ENV: "test"
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="rule")
def fixture_rule():
    return RuleJobEnvironmentPrefix()


def test_rule_on_correct_workflow(rule, correct_workflow):
    obj = correct_workflow.jobs["job-key"]

    result, message = rule.fn(correct_workflow.jobs["job-key"])
    assert result is True
    assert message == ""

    finding = rule.execute(obj)
    assert finding is None


def test_rule_on_no_env_workflow(rule, no_env_workflow):
    obj = no_env_workflow.jobs["job-key"]

    result, message = rule.fn(no_env_workflow.jobs["job-key"])
    assert result is True
    assert message == ""

    finding = rule.execute(obj)
    assert finding is None


def test_rule_on_missing_prefix_workflow(rule, missing_prefix_workflow):
    obj = missing_prefix_workflow.jobs["job-key"]

    result, message = rule.fn(obj)
    assert result is False
    assert "TEST_ENV" in message

    finding = rule.execute(obj)
    assert "TEST_ENV" in finding.description


def test_fail_compatibility(rule, correct_workflow):
    finding = rule.execute(correct_workflow)
    assert "Workflow not compatible with" in finding.description

    finding = rule.execute(correct_workflow.jobs["job-key"].steps[0])
    assert "Step not compatible with" in finding.description



"""Tests src/bitwarden_workflow_linter/rule.py."""

import pytest
from typing import Union

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.models.job import Job
from src.bitwarden_workflow_linter.models.step import Step
from src.bitwarden_workflow_linter.models.workflow import Workflow
from src.bitwarden_workflow_linter.rule import Rule, RuleExecutionException


yaml = YAML()


@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
name: Test Workflow

on:
  workflow_dispatch:

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - name: Test
        uses: actions/checkout@main
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="incorrect_workflow")
def fixture_incorrect_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@main
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


class RuleStep(Rule):
    def __init__(self):
        self.message = "test"
        self.on_fail = "error"
        self.compatibility = [Step]


class RuleNameExists(Rule):
    def __init__(self):
        self.message = "name must exist"
        self.on_fail = "error"

    def fn(self, obj: Union[Workflow, Job, Step]) -> bool:
        print(f"{type(self).__name__}\n{obj}")
        return obj.name is not None, self.message


class RuleException(Rule):
    def __init__(self):
        self.message = "should raise Exception"
        self.on_fail = "error"

    def fn(self, obj: Union[Workflow, Job, Step]) -> bool:
        raise RuleExecutionException("test Exception")


@pytest.fixture(name="step_rule")
def fixture_step_rule():
    return RuleStep()


@pytest.fixture(name="exists_rule")
def fixture_exists_rule():
    return RuleNameExists()


@pytest.fixture(name="exception_rule")
def fixture_exception_rule():
    return RuleException()


def test_build_lint_message(step_rule, correct_workflow):
    assert step_rule.build_lint_message("test", correct_workflow) == "Workflow => test"

    assert (
        step_rule.build_lint_message("test", correct_workflow.jobs["job-key"])
        == "Job [job-key] => test"
    )

    assert (
        step_rule.build_lint_message("test", correct_workflow.jobs["job-key"].steps[0])
        == "Step [job-key.0] => test"
    )


def test_rule_compatibility(step_rule, correct_workflow):
    assert "not compatible" in step_rule.execute(correct_workflow).description
    assert (
        "not compatible"
        in step_rule.execute(correct_workflow.jobs["job-key"]).description
    )
    assert (
        "not compatible"
        not in step_rule.execute(correct_workflow.jobs["job-key"].steps[0]).description
    )


def test_correct_rule_execution(exists_rule, correct_workflow):
    assert exists_rule.execute(correct_workflow) is None
    assert exists_rule.execute(correct_workflow.jobs["job-key"]) is None
    assert exists_rule.execute(correct_workflow.jobs["job-key"].steps[0]) is None


def test_incorrect_rule_execution(exists_rule, incorrect_workflow):
    assert "name must exist" in exists_rule.execute(incorrect_workflow).description
    assert (
        "name must exist"
        in exists_rule.execute(incorrect_workflow.jobs["job-key"]).description
    )
    assert (
        "name must exist"
        in exists_rule.execute(incorrect_workflow.jobs["job-key"].steps[0]).description
    )


def test_exception_rule_execution(exception_rule, incorrect_workflow):
    assert "failed to apply" in exception_rule.execute(incorrect_workflow).description



"""Test src/bitwarden_workflow_linter/rules/name_capitalized.py."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.name_capitalized import RuleNameCapitalized

yaml = YAML()


@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
name: Test Workflow

on:
  workflow_dispatch:

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - name: Test
        run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="incorrect_workflow")
def fixture_incorrect_workflow():
    workflow = """\
---
name: test
on:
  workflow_dispatch:

jobs:
  job-key:
    name: test
    runs-on: ubuntu-latest
    steps:
      - name: test
        run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="underscore_name_workflow")
def fixture_underscore_name_workflow():
    workflow = """\
name: _test
on:
  workflow_dispatch:

jobs:
  job-key:
    name: test
    runs-on: ubuntu-latest
    steps:
      - name: test
        run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="missing_name_workflow")
def fixture_missing_name_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-latest
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="rule")
def fixture_rule():
    return RuleNameCapitalized()


def test_rule_on_correct_workflow(rule, correct_workflow):
    result, _ = rule.fn(correct_workflow)
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[0])
    assert result is True


def test_rule_on_incorrect_workflow_name(rule, incorrect_workflow):
    result, _ = rule.fn(incorrect_workflow)
    assert result is False


def test_rule_on_underscore_name_workflow(rule, underscore_name_workflow):
    result, _ = rule.fn(underscore_name_workflow)
    assert result is True


def test_rule_on_incorrect_job_name(rule, incorrect_workflow):
    result, _ = rule.fn(incorrect_workflow.jobs["job-key"])
    assert result is False


def test_rule_on_incorrect_step_name(rule, incorrect_workflow):
    result, _ = rule.fn(incorrect_workflow.jobs["job-key"].steps[0])
    assert result is False


def test_rule_on_missing_names(rule, missing_name_workflow):
    result, _ = rule.fn(missing_name_workflow)
    assert result is True

    result, _ = rule.fn(missing_name_workflow.jobs["job-key"])
    assert result is True

    result, _ = rule.fn(missing_name_workflow.jobs["job-key"].steps[0])
    assert result is True



"""A Rule to enforce check-run is run when workflow uses pull_request_target."""

from typing import Optional, Tuple

from ..models.workflow import Workflow
from ..rule import Rule
from ..utils import LintLevels, Settings


class RuleCheckPrTarget(Rule):
    def __init__(self, settings: Optional[Settings] = None, lint_level: Optional[LintLevels] = LintLevels.NONE) -> None:
        """
        To ensure pull_request_target is safe to use, the check-run step is added
        to all jobs as a dependency.

        Once a branch is pushed to Github, it already opens up a vulnerability
        even if the check-run scan fails to detect this.

        In order to prevent a vulnerable branch from being used for an attack
        prior to being caught through vetting, all pull_request_target workflows
        should only be run by users with appropriate permissions.

        This greatly reduces the risk as community contributors can't use a fork to run a compromised workflow that uses pull_request_target.
        """
        self.message = "A check-run job must be included as a direct job dependency when pull_request_target is used and the workflow may only apply to runs on the main branch"
        self.on_fail = lint_level
        self.compatibility = [Workflow]
        self.settings = settings

    def targets_main_branch(self, obj: Workflow) -> bool:
        default_branch = self.settings.default_branch
        branches = obj.on["pull_request_target"].get("branches", [])
        if isinstance(branches, str):
            branches = [branches]
        return len(branches) == 1 and branches[0] == default_branch

    def has_check_run(self, obj: Workflow) -> Tuple[bool, str]:
        for name, job in obj.jobs.items():
            if job.uses == "bitwarden/gh-actions/.github/workflows/check-run.yml@main":
                return True, name
        return False, ""

    def check_run_required(self, obj:Workflow, check_job:str) -> list:
        missing_jobs = []
        for job in list(obj.jobs.keys()):
            if job is not check_job:
                job_list = obj.jobs[job].needs
                if job_list:
                    if check_job not in job_list:
                        missing_jobs.append(job)
                else:
                    missing_jobs.append(job)
        return missing_jobs

    def fn(self, obj: Workflow) -> Tuple[bool, str]:
        Errors = []
        if obj.on.get("pull_request_target"):
            result, check_job = self.has_check_run(obj)
            main_branch_only = self.targets_main_branch(obj)
            if not main_branch_only:
                default_branch = self.settings.default_branch
                Errors.append(f"Workflows using pull_request_target can only target the '{default_branch}' branch")
            if result:
                missing_jobs = self.check_run_required(obj, check_job)
                if missing_jobs:
                    job_list = ', '.join(missing_jobs)
                    message = f"Check-run is missing from the following jobs in the workflow: {job_list}"
                    Errors.append(message)
            else:
                message = f"A check-run job must be included as a direct job dependency when pull_request_target is used"
                Errors.append(message)
            if Errors:
                self.message = "\n".join(Errors)
                return False, f"{self.message}"
            else:
                return True, ""
        else:
            return True, ""



"""A Rule to enforce all 'name' values start with a capital letter."""

import re
from typing import Optional, Tuple, Union

from ..models.job import Job
from ..models.step import Step
from ..models.workflow import Workflow
from ..rule import Rule
from ..utils import LintLevels, Settings


class RuleNameCapitalized(Rule):
    """Rule to enforce all 'name' values start with a capital letter.

    A simple standard to help keep uniformity in naming.
    """

    def __init__(self, settings: Optional[Settings] = None, lint_level: Optional[LintLevels] = LintLevels.NONE) -> None:
        """Constructor for RuleNameCapitalized to override the Rule class.

        Args:
          settings:
            A Settings object that contains any default, overridden, or custom settings
            required anywhere in the application.
          lint_level:
            The LintLevels enum value to determine the severity of the rule.
        """
        self.message = "name must capitalized"
        self.on_fail = lint_level
        self.settings = settings

    def fn(self, obj: Union[Workflow, Job, Step]) -> Tuple[bool, str]:
        """Enforces capitalization of the first letter of any name key.

        Example:
        ---
        name: Test Workflow

        on:
          workflow_dispatch:

        jobs:
          job-key:
            name: Test
            runs-on: ubuntu-latest
            steps:
              - name: Test
                run: echo test

        'Test Workflow', 'Test', and 'Test' all start with a capital letter.

        See tests/rules/test_name_capitalized.py for examples of incorrectly
        capitalized names. This Rule DOES NOT enforce that the name exists.
        It only enforces capitalization IF it does.
        """
        if isinstance(obj, Workflow):
            if obj.name:
                if obj.name[0] != "_":
                    return obj.name[0].isupper(), self.message
        else:
            if obj.name and not re.match(r"^\s*\${{\s*matrix\..*}}.*", obj.name):
                return obj.name[0].isupper(), self.message

        return True, ""  # Force passing



"""Test src/bitwarden_workflow_linter/models/workflow.py."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.models.workflow import Workflow


yaml = YAML()


@pytest.fixture(name="simple_workflow_yaml")
def fixture_simple_workflow_yaml():
    return yaml.load(
        """\
---
name: test
on:
  workflow_dispatch:

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - name: Test
        run: echo test
"""
    )


@pytest.fixture(name="complex_workflow_yaml")
def fixture_complex_workflow_yaml():
    return yaml.load(
        """\
---
name: test
on:
  workflow_dispatch:
  pull_request:

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - name: Test
        run: echo test

  call-workflow:
    uses: bitwarden/server/.github/workflows/workflow-linter.yml@master

  test-normal-action:
    name: Download Latest
    runs-on: ubuntu-20.04
    steps:
      - name: Checkout
        uses: actions/checkout@2541b1294d2704b0964813337f33b291d3f8596b

      - run: |
          echo test

  test-local-action:
    name: Testing a local action call
    runs-on: ubuntu-20.04
    steps:
      - name: local-action
        uses: ./version-bump
"""
    )


def test_simple_workflow(simple_workflow_yaml):
    workflow = Workflow.init("", None, simple_workflow_yaml)

    assert workflow.name == "test"
    assert len(workflow.on.keys()) == 1
    assert len(workflow.jobs.keys()) == 1


def test_complex_workflow(complex_workflow_yaml):
    workflow = Workflow.init("", None, complex_workflow_yaml)

    assert workflow.name == "test"
    assert len(workflow.on.keys()) == 2
    assert len(workflow.jobs.keys()) == 4


def test_workflow_extra_kwargs(simple_workflow_yaml):
    extra_data_workflow = simple_workflow_yaml
    extra_data_workflow["extra"] = "This should not exist"

    workflow = Workflow.init("", None, extra_data_workflow)

    with pytest.raises(Exception):
        assert workflow.extra == "test"



"""Test src/bitwarden_workflow_linter/rules/check_pr_target."""

import pytest

from ruamel.yaml import YAML
from unittest.mock import patch

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.check_pr_target import (
    RuleCheckPrTarget,
)
from src.bitwarden_workflow_linter.models.workflow import Workflow
from src.bitwarden_workflow_linter.utils import Settings

yaml = YAML()
message = "A check-run job must be included as a direct job dependency when pull_request_target is used and the workflow may only apply to runs on the main branch"

@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]
    branches:
      - 'main'

jobs:
  check-run:
    name: Check PR run
    uses: bitwarden/gh-actions/.github/workflows/check-run.yml@main

  quality:
    name: Quality scan
    needs: check-run
    steps:
      - run: echo test

  dependent-job:
     name: Another Dependent Job
     needs:
       - quality
       - check-run
     steps:
       - run: echo another dependent job
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="no_check_workflow")
def fixture_no_check_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]
    branches:
      - 'main'

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="no_needs_workflow")
def fixture_no_needs_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]
    branches: main

jobs:
  check-run:
    name: Check PR run
    uses: bitwarden/gh-actions/.github/workflows/check-run.yml@main

  job-key:
    runs-on: ubuntu-22.04
    steps:
      - run: echo test

  quality:
    name: Quality scan
    steps:
      - run: echo test

"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="no_target_workflow")
def fixture_no_target_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="dependent_missing_check_workflow")
def dependent_missing_check_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]
    branches:
      - 'main'

jobs:
  check-run:
    name: Check PR run
    uses: bitwarden/gh-actions/.github/workflows/check-run.yml@main

  quality:
    name: Quality scan
    needs: check-run
    steps:
      - run: echo test

  dependent-job:
     name: Another Dependent Job
     needs:
       - quality
     steps:
       - run: echo another dependent job
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="no_branches_workflow")
def fixture_no_branches_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]

jobs:
  check-run:
    name: Check PR run
    uses: bitwarden/gh-actions/.github/workflows/check-run.yml@main

  quality:
    name: Quality scan
    needs: check-run
    steps:
      - run: echo test

  dependent-job:
     name: Another Dependent Job
     needs:
       - quality
       - check-run
     steps:
       - run: echo another dependent job
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="bad_branches_workflow")
def fixture_bad_branches_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]
    branches:
      - 'main'
      - 'not main'

jobs:
  check-run:
    name: Check PR run
    uses: bitwarden/gh-actions/.github/workflows/check-run.yml@main

  quality:
    name: Quality scan
    needs: check-run
    steps:
      - run: echo test

  dependent-job:
     name: Another Dependent Job
     needs:
       - quality
       - check-run
     steps:
       - run: echo another dependent job
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="two_failures_workflow")
def fixture_two_failures_workflow():
    workflow = """\
---
on:
  workflow_dispatch:
  pull_request_target:
    types: [opened, synchronize]

jobs:
  check-run:
    name: Check PR run
    uses: bitwarden/some-other-repo/.github/workflows/check-run.yml@main

  quality:
    name: Quality scan
    needs: check-run
    steps:
      - run: echo test

  dependent-job:
     name: Another Dependent Job
     needs:
       - quality
       - check-run
     steps:
       - run: echo another dependent job
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)

@pytest.fixture(name="rule")
def fixture_rule():
  settings= Settings(default_branch="main")
  return RuleCheckPrTarget(settings=settings)

@pytest.fixture
def mock_workflow():
    return Workflow(
        on={
            "pull_request_target": {
                "branches": ["main"]
            }
        },
        jobs={}
    )
def test_targets_main_branch_with_settings_yaml(mock_workflow):
    settings = Settings(default_branch="main")
    rule = RuleCheckPrTarget(settings=settings)
    assert rule.targets_main_branch(mock_workflow) is True

def test_targets_custom_default_branch(mock_workflow):
    settings = Settings(
        enabled_rules=[],
        approved_actions={},
        actionlint_version="",
        default_branch="production"
    )
    rule = RuleCheckPrTarget(settings=settings)
    mock_workflow.on["pull_request_target"]["branches"] = ["production"]
    assert rule.targets_main_branch(mock_workflow) is True

def test_targets_main_branch_no_default_branch(mock_workflow):
    # Mock the factory method to simulate loading default settings
    with patch("src.bitwarden_workflow_linter.utils.Settings.factory") as mock_factory:
        # Simulate the default settings returned by the factory
        mock_factory.return_value = Settings(default_branch="main")

        # Use the mocked factory to create the Settings instance
        settings = Settings.factory()
        rule = RuleCheckPrTarget(settings=settings)

        # Assert that the workflow targets the main branch
        assert rule.targets_main_branch(mock_workflow) is True

def test_targets_main_branch_incorrect_branch(mock_workflow):
    settings = Settings({"default_branch": "main"})
    rule = RuleCheckPrTarget(settings=settings)
    mock_workflow.on["pull_request_target"]["branches"] = ["feature"]
    assert rule.targets_main_branch(mock_workflow) is False

def test_targets_main_branch_no_branches_specified(mock_workflow):
    settings = Settings({"default_branch": "main"})
    rule = RuleCheckPrTarget(settings=settings)
    mock_workflow.on["pull_request_target"].pop("branches")
    assert rule.targets_main_branch(mock_workflow) is False

def test_rule_on_correct_workflow(rule, correct_workflow):
    result, message = rule.fn(correct_workflow)
    assert result is True
    assert message == ""

def test_rule_on_no_checkworkflow(rule, no_check_workflow):
    result, message = rule.fn(no_check_workflow)
    assert result is False
    assert message == message

def test_rule_on_no_target_workflow(rule, no_target_workflow):
    result, message = rule.fn(no_target_workflow)
    assert result is True
    assert message == ""

def test_rule_on_jobs_without_needs(rule, no_needs_workflow):
    result, message = rule.fn(no_needs_workflow)
    assert result is False
    assert message == message, "check-run is missing from the following jobs in the workflow: quality"

def test_rule_on_dependencies_without_check(rule, dependent_missing_check_workflow):
    result, message = rule.fn(dependent_missing_check_workflow)
    assert result is False
    assert message == message, "check-run is missing from the following jobs in the workflow: dependent-job"

def test_rule_on_no_branches_workflow(rule, no_branches_workflow):
    result, message = rule.fn(no_branches_workflow)
    assert result is False
    assert message == "Workflows using pull_request_target can only target the 'main' branch"

def test_rule_on_only_target_main(rule, bad_branches_workflow):
    result, message = rule.fn(bad_branches_workflow)
    assert result is False
    assert message == "Workflows using pull_request_target can only target the 'main' branch"

def test_rule_on_two_failures(rule, two_failures_workflow):
    result, message = rule.fn(two_failures_workflow)
    assert result is False
    assert message == "Workflows using pull_request_target can only target the 'main' branch\nA check-run job must be included as a direct job dependency when pull_request_target is used"



"""Test src/bitwarden_workflow_linter/rules/step_pinned.py."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.step_pinned import RuleStepUsesPinned

yaml = YAML()


@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - name: Test 3rd Party Action
        uses: actions/checkout@b4ffde65f46336ab88eb53be808477a3936bae11 # v4.1.1

      - name: Test Internal Action
        uses: bitwarden/ios/.github/actions/dispatch-and-download@main

      - name: Test Local Action
        uses: ./actions/test-action

      - name: Test Run Action
        run: echo "test"

      - name: Test Bitwarden Action
        uses: bitwarden/sm-action@92d1d6a4f26a89a8191c83ab531a53544578f182 # v2.0.0
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="incorrect_workflow")
def fixture_incorrect_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - name: Test 3rd Party Action without version pinned
        uses: actions/checkout@b4ffde65f46336ab88eb53be808477a3936bae11

      - name: Test External Branch
        uses: actions/checkout@main

      - name: Test Incorrect Hex
        uses: actions/checkout@b4ffde

      - name: Test Internal Commit
        uses: bitwarden/gh-actions/get-keyvault-secrets@b4ffde65f46336ab88eb53be808477a3936bae11 # v4.1.1
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="rule")
def fixture_rule():
    return RuleStepUsesPinned()


def test_rule_on_correct_workflow(rule, correct_workflow):
    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[0])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[1])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[2])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[3])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[4])
    assert result is True


def test_rule_on_incorrect_workflow_version_comment(rule, incorrect_workflow):
    result, message = rule.fn(incorrect_workflow.jobs["job-key"].steps[0])
    assert result is False
    assert "the version of the action commit sha" in message


def test_rule_on_incorrect_workflow_external_branch(rule, incorrect_workflow):
    result, message = rule.fn(incorrect_workflow.jobs["job-key"].steps[1])
    assert result is False
    assert "Please pin the action" in message


def test_rule_on_incorrect_workflow_hex(rule, incorrect_workflow):
    result, message = rule.fn(incorrect_workflow.jobs["job-key"].steps[2])
    assert result is False
    assert "Please use the full commit sha" in message


def test_rule_on_incorrect_workflow_internal_commit(rule, incorrect_workflow):
    result, message = rule.fn(incorrect_workflow.jobs["job-key"].steps[3])
    assert result is False
    assert "Please pin to main" in message


def test_fail_compatibility(rule, correct_workflow):
    finding = rule.execute(correct_workflow)
    assert "Workflow not compatible with" in finding.description

    finding = rule.execute(correct_workflow.jobs["job-key"])
    assert "Job not compatible with" in finding.description



"""Tests src/bitwarden_workflow_linter/utils.py."""

from src.bitwarden_workflow_linter.utils import Action, Colors, LintFinding, LintLevels


def test_action_eq():
    action_def = {"name": "bitwarden/sm-action", "version": "1.0.0", "sha": "some-sha"}

    action_a = Action(**action_def)
    action_b = Action(**action_def)

    assert (action_a == action_b) is True
    assert (action_a != action_b) is False


def test_action_ne():
    action_a = Action(name="bitwarden/sm-action", version="1.0.0", sha="some-sha")
    action_b = Action(name="bitwarden/sm-action", version="1.1.0", sha="some-other-sha")

    assert (action_a == action_b) is False
    assert (action_a != action_b) is True


def test_lint_level():
    warning = LintLevels.WARNING
    assert warning.code == 1
    assert warning.color == Colors.yellow


def test_lint_finding():
    warning = LintFinding(description="<no description>", level=LintLevels.WARNING)
    assert str(warning) == "\x1b[33mwarning\x1b[0m <no description>"

    error = LintFinding(description="<no description>", level=LintLevels.ERROR)
    assert str(error) == "\x1b[31merror\x1b[0m <no description>"



"""A Rule to enforce pinning runners to a specific OS version."""

from typing import Optional, Tuple

from ..models.job import Job
from ..rule import Rule
from ..utils import LintLevels, Settings


class RuleJobRunnerVersionPinned(Rule):
    """Rule to enforce pinned Runner OS versions.

    Using `*-latest` versions will update automatically and has broken all of
    our workflows in the past. To avoid this and prevent a single event from
    breaking the majority of our pipelines, we pin the versions.
    """

    def __init__(self, settings: Optional[Settings] = None, lint_level: Optional[LintLevels] = LintLevels.NONE) -> None:
        """Constructor for RuleJobRunnerVersionPinned to override Rule class.

        Args:
          settings:
            A Settings object that contains any default, overridden, or custom settings
            required anywhere in the application.
          lint_level:
            The LintLevels enum value to determine the severity of the rule.
        """
        self.message = "Workflow runner must be pinned"
        self.on_fail = lint_level
        self.compatibility = [Job]
        self.settings = settings

    def fn(self, obj: Job) -> Tuple[bool, str]:
        """Enforces runners are pinned to a version

        Example:
        ---
        on:
          workflow_dispatch:

        jobs:
          job-key:
            runs-on: ubuntu-22.04
            steps:
              - run: echo test

        call-workflow:
          uses: bitwarden/server/.github/workflows/workflow-linter.yml@master

        'runs-on' is pinned to '22.04' instead of 'latest'
        """
        if obj.runs_on is not None and "latest" in obj.runs_on:
            return False, self.message
        return True, ""



"""A Rule to run zizmor on workflows."""

from typing import Optional, Tuple
import subprocess
import platform
import urllib.request
import tempfile
import os

from ..rule import Rule
from ..models.workflow import Workflow
from ..utils import LintLevels, Settings


def install_zizmor(platform_system: str, version: str) -> Tuple[bool, str]:
    """Install zizmor via pip."""
    error = f"An error occurred when installing Zizmor on {platform_system}"
    try:
        subprocess.run(
            ["pip", "install", f"zizmor=={version}"], check=True, capture_output=True
        )
        return True, ""
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False, f"{error} : check pip installation"


def check_zizmor_path(platform_system: str, version: str) -> Tuple[bool, str]:
    """Check if zizmor is in the system's PATH."""
    try:
        result = subprocess.run(
            ["zizmor", "--version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            text=True,
        )
        installed_version = result.stdout.strip().split()[-1] if result.stdout else ""
        if version in installed_version:
            return True, ""
        else:
            return install_zizmor(platform_system, version)
    except subprocess.CalledProcessError:
        return (
            False,
            (
                "Failed to install zizmor, please check your pip "
                "installation or manually install it"
            ),
        )
    except FileNotFoundError:
        return install_zizmor(platform_system, version)


def download_config_file(config_url: str) -> Optional[str]:
    """Download zizmor config file from remote URL."""
    if not config_url:
        return None

    try:
        with urllib.request.urlopen(config_url) as response:
            config_content = response.read()

        # Create temporary file for config
        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".yml", delete=False
        ) as temp_file:
            temp_file.write(config_content.decode("utf-8"))
            return temp_file.name
    except (urllib.error.URLError, IOError):
        return None


class RunZizmor(Rule):
    """Rule to run zizmor as part of workflow linter."""

    def __init__(
        self,
        settings: Optional[Settings] = None,
        lint_level: Optional[LintLevels] = LintLevels.WARNING,
    ) -> None:
        self.message = "Zizmor must pass without errors"
        self.on_fail = lint_level
        self.compatibility = [Workflow]
        self.settings = settings

    def fn(self, obj: Workflow) -> Tuple[bool, str]:
        if not obj or not obj.filename:
            raise AttributeError(
                "Running zizmor without a filename is not currently supported"
            )

        if not self.settings.zizmor_version:
            raise KeyError("The 'zizmor_version' is missing in the configuration file.")

        # Check if zizmor is already installed
        installed, error = check_zizmor_path(
            platform.system(), self.settings.zizmor_version
        )
        if not installed:
            return False, error

        # Build zizmor command
        cmd = ["zizmor", "--format", "plain"]

        # Add config file if specified
        config_file = None
        if self.settings.zizmor_config_url:
            config_file = download_config_file(self.settings.zizmor_config_url)
            if config_file:
                cmd.extend(["--config", config_file])

        # Add the workflow file
        cmd.append(obj.filename)

        try:
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
            )

            # Clean up temporary config file if created
            if config_file:
                try:
                    os.unlink(config_file)
                except OSError:
                    pass

            # zizmor returns 0 for success, non-zero for findings or errors
            if result.returncode == 0:
                return True, ""
            else:
                # Return the findings/errors from zizmor
                output = result.stdout if result.stdout else result.stderr
                return False, output

        except (subprocess.CalledProcessError, FileNotFoundError, OSError) as e:
            # Clean up temporary config file if created
            if config_file:
                try:
                    os.unlink(config_file)
                except OSError:
                    pass
            return False, f"Error running zizmor: {str(e)}"



"""This is the entrypoint module for the workflow-linter CLI."""

import argparse
import sys

from typing import List, Optional

from .actions import ActionsCmd
from .lint import LinterCmd
from .utils import Settings

local_settings = Settings.factory()


def main(input_args: Optional[List[str]] = None) -> int:
    """CLI utility to lint GitHub Action Workflows.

    A CLI utility to enforce coding standards on GitHub Action workflows. The
    utility also provides other subcommands to assist with other workflow
    maintenance tasks; such as maintaining the list of approved GitHub Actions.
    """
    linter_cmd = LinterCmd(settings=local_settings)
    actions_cmd = ActionsCmd(settings=local_settings)

    # Read arguments from command line.
    parser = argparse.ArgumentParser(prog="bwwl")
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    subparsers = parser.add_subparsers(required=True, dest="command")

    subparsers = LinterCmd.extend_parser(subparsers)
    subparsers = ActionsCmd.extend_parser(subparsers)

    # Pull the arguments from the command line
    input_args = sys.argv[1:]
    if not input_args:
        raise SystemExit(parser.print_help())

    args = parser.parse_args(input_args)
    if args.command == "lint":
        if len(args.files) > 1:
            raise parser.error("Only one -f / --files flag must be specified")
        return linter_cmd.run(args.files[0], args.strict)

    if args.command == "actions":
        print(f"{'-'*50}\n!!bwwl actions is in BETA!!\n{'-'*50}")
        if args.actions_command == "add":
            return actions_cmd.add(args.name, args.output)
        if args.actions_command == "update":
            return actions_cmd.update(args.output)

    return -1


if __name__ == "__main__":
    sys.exit(main())



"""A Rule to enforce the use of a list of pre-approved Actions."""

from typing import Optional, Tuple

from ..models.step import Step
from ..rule import Rule
from ..utils import LintLevels, Settings


class RuleStepUsesApproved(Rule):
    """Rule to enforce that all Actions have been pre-approved.

    To limit the surface area of a supply chain attack in our pipelines, all Actions
    are required to pass a security review and be added to the pre-approved list to
    check against.
    """

    def __init__(self, settings: Optional[Settings] = None, lint_level: Optional[LintLevels] = LintLevels.NONE) -> None:
        """Constructor for RuleStepUsesApproved to override Rule class.

        Args:
          settings:
            A Settings object that contains any default, overridden, or custom settings
            required anywhere in the application.
          lint_level:
            The LintLevels enum value to determine the severity of the rule.
        """
        self.on_fail = lint_level
        self.compatibility = [Step]
        self.settings = settings

    def skip(self, obj: Step) -> bool:
        """Skip this Rule on some Steps.

        This Rule does not apply to a few types of Steps. These
        Rules are skipped.
        """
        ## Force pass for any shell steps
        if not obj.uses:
            return True

        ## Force pass for any local actions
        if "@" not in obj.uses:
            return True

        ## Force pass for any bitwarden/
        if obj.uses.startswith("bitwarden/"):
            return True

        return False

    def fn(self, obj: Step) -> Tuple[bool, str]:
        """Enforces all externally used Actions are on the pre-approved list.

        The pre-approved list allows tight auditing on what Actions are trusted
        and allowed to be run in our environments. This helps mitigate risks
        against supply chain attacks in our pipelines.

        Example:
        ---
        on:
          workflow_dispatch:

        jobs:
          job-key:
            runs-on: ubuntu-22.04
            steps:
              - name: Checkout Branch
                uses: actions/checkout@b4ffde65f46336ab88eb53be808477a3936bae11 # v4.1.1

              - name: Test Bitwarden Action
                uses: bitwarden/gh-actions/get-keyvault-secrets@main

              - name: Test Local Action
                uses: ./actions/test-action

              - name: Test Run Action
                run: echo "test"

        In this example, 'actions/checkout' must be on the pre-approved list
        and the metadata must match in order to succeed. The other three
        Steps will be skipped.
        """
        if self.skip(obj):
            return True, ""

        obj_path = obj.uses_path

        # Remove the action directory if the action is in a multi-actions repo
        if len(obj.uses_path.split("/")) > 2:
            obj_path = "/".join(obj.uses_path.split("/")[:-1])

        # Actions in bitwarden/ are auto-approved
        if obj.uses and not obj_path in self.settings.approved_actions:
            return False, (
                f"New Action detected: {obj.uses_path}\nFor security purposes, "
                "actions must be reviewed and be on the pre-approved list"
            )

        return True, ""



"""Test src/bitwarden_workflow_linter/rules/step_approved.py."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.step_approved import RuleStepUsesApproved
from src.bitwarden_workflow_linter.utils import Settings


yaml = YAML()


@pytest.fixture(name="settings")
def fixture_settings():
    return Settings(
        approved_actions={
            "actions/checkout": {
                "name": "actions/checkout",
                "version": "v4.1.1",
                "sha": "b4ffde65f46336ab88eb53be808477a3936bae11",
            },
            "actions/download-artifact": {
                "name": "actions/download-artifact",
                "version": "v4.1.0",
                "sha": "f44cd7b40bfd40b6aa1cc1b9b5b7bf03d3c67110",
            },
        }
    )


@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - name: Checkout Branch
        uses: actions/checkout@b4ffde65f46336ab88eb53be808477a3936bae11 # v4.1.1

      - name: Test Bitwarden Action
        uses: bitwarden/gh-actions/get-keyvault-secrets@main

      - name: Test Local Action
        uses: ./actions/test-action

      - name: Test Run Action
        run: echo "test"
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="incorrect_workflow")
def fixture_incorrect_workflow():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - name: Checkout Branch
        uses: joseph-flinn/action-DNE@main

      - name: Checkout Branch with Ref
        uses: notbitwarden/subfolder/action-DNE@main
        with:
            ref: main

"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="rule")
def fixture_rule(settings):
    return RuleStepUsesApproved(settings=settings)


def test_rule_on_correct_workflow(rule, correct_workflow):
    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[0])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[1])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[2])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[3])
    assert result is True


def test_rule_on_incorrect_workflow(rule, incorrect_workflow):
    result, message = rule.fn(incorrect_workflow.jobs["job-key"].steps[0])
    assert result is False
    assert "New Action detected" in message

    result, message = rule.fn(incorrect_workflow.jobs["job-key"].steps[1])
    assert result is False
    assert "New Action detected" in message


def test_fail_compatibility(rule, correct_workflow):
    finding = rule.execute(correct_workflow)
    assert "Workflow not compatible with" in finding.description

    finding = rule.execute(correct_workflow.jobs["job-key"])
    assert "Job not compatible with" in finding.description



"""Test src/bitwarden_workflow_linter/rules/underscore_outputs.py."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.underscore_outputs import RuleUnderscoreOutputs

yaml = YAML()


@pytest.fixture(name="correct_workflow")
def fixture_correct_workflow():
    workflow = """\
---
name: Test Correct Workflow
on:
  workflow_dispatch:
    outputs:
      registry:
        value: 'Test Value'
      some_registry:
        value: 'Test Value'
  workflow_call:
    outputs:
      registry:
        value: 'Test Value'
      some_registry:
        value: 'Test Value'
  push: {}

jobs:
  job-key:
    name: Test Correct Job
    runs-on: ubuntu-22.04
    outputs:
      test_key_job: ${{ steps.test_output_1.outputs.test_key }}
    steps:
      - name: Test output in one-line run step
        id: test_output_1
        run: echo "test_key_1=Test-Value1" >> $GITHUB_OUTPUT

      - name: Test output in multi-line run step
        id: test_output_2
        run: |
          echo
          fake-command=Test-Value2
          echo "test_key_2=$REF" >> $GITHUB_OUTPUT
          echo "deployed_ref=$DEPLOYED_REF" >> $GITHUB_OUTPUT

      - name: Test step with one-line run and no Output
        id: test_output_3
        run: echo "test_key_3"

      - name: Test step with multi-line run and no Output
        id: test_output_4
        run: |
          echo
          fake-command=Test-Value4
          echo "test_key_4"
          echo "deployed_ref"

      - name: Test step with no run
        id: test_output_5
        uses: actions/checkout@v2
        with:
          ref: ${{ github.ref }}
          fetch-depth: 0
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="incorrect_workflow")
def fixture_incorrect_workflow():
    workflow = """\
---
name: Test Incorrect Workflow
on:
  workflow_dispatch:
    outputs:
      registry-1:
        value: 'Test Value'
      some_registry-1:
        value: 'Test Value'
  workflow_call:
    outputs:
      registry-2:
        value: 'Test Value'
  push: {}

jobs:
  job-key:
    name: Test Incorrect Job
    runs-on: ubuntu-22.04
    outputs:
      test-key-1: ${{ steps.test_output_1.outputs.test_key }}
    steps:
      - name: Test step with one-line run and no Output
        id: test_output_3
        run: echo "test-key-3"

      - name: Test step with multi-line run and no Output
        id: test_output_4
        run: |
          echo
          fake-command=Test-Value4
          echo "test-key-4"
          echo "deployed-ref"
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="push_only_workflow")
def fixture_push_only_workflow():
    workflow = """\
---
name: Push Only
on:
  push: {}

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - run: echo test

"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="misc_workflow")
def fixture_misc_workflow():
    workflow = """\
---
name: Misc Workflow
on:
  push: {}

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - name: Test Step
        run: |
          echo "test_value=$does_it_break" >> /tmp/test

      - name: Test Step 2
        run: |
          TEST_FILE=/tmp/test2
          echo "test_value=$does_it_break" >> $TEST_FILE
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="no_output_workflow")
def fixture_no_output_workflow():
    workflow = """\
---
name: No Output Workflow
on:
  workflow_dispatch:
    inputs: {}
  workflow_call: {}

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="rule")
def fixture_rule():
    return RuleUnderscoreOutputs()


def test_rule_on_correct_workflow(rule, correct_workflow):
    result, _ = rule.fn(correct_workflow)
    assert result is True


def test_rule_on_incorrect_workflow(rule, incorrect_workflow):
    result, _ = rule.fn(incorrect_workflow)
    assert result is False


def test_rule_on_correct_job(rule, correct_workflow):
    result, _ = rule.fn(correct_workflow.jobs["job-key"])
    assert result is True


def test_rule_on_incorrect_job(rule, incorrect_workflow):
    result, _ = rule.fn(incorrect_workflow.jobs["job-key"])
    assert result is False


def test_rule_on_correct_step(rule, correct_workflow):
    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[0])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[1])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[2])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[3])
    assert result is True

    result, _ = rule.fn(correct_workflow.jobs["job-key"].steps[4])
    assert result is True


def test_rule_on_incorrect_step(rule, incorrect_workflow):
    result, _ = rule.fn(incorrect_workflow.jobs["job-key"].steps[0])
    assert result is True

    result, _ = rule.fn(incorrect_workflow.jobs["job-key"].steps[1])
    assert result is True


def test_rule_on_push_only_workflow(rule, push_only_workflow):
    result, _ = rule.fn(push_only_workflow)
    assert result is True


def test_rule_on_misc_workflow(rule, misc_workflow):
    result, _ = rule.fn(misc_workflow.jobs["job-key"].steps[0])
    assert result is True

    result, _ = rule.fn(misc_workflow.jobs["job-key"].steps[1])
    assert result is True


def test_rule_on_no_output_workflow(rule, no_output_workflow):
    result, _ = rule.fn(no_output_workflow)
    assert result is True



"""A Rule to run actionlint on workflows."""

from typing import Optional, Tuple
import subprocess
import platform
import urllib.request
import os

from ..rule import Rule
from ..models.workflow import Workflow
from ..utils import LintLevels, Settings


def install_actionlint(platform_system: str, version: str) -> Tuple[bool, str]:
    """If actionlint is not installed, detects OS platform
    and installs actionlint"""

    error = f"An error occurred when installing Actionlint on {platform_system}"

    if platform_system.startswith(("Linux", "Darwin")):
        """Homebrew does not maintain non-latest versions of actionlint so Mac will install from source"""
        return install_actionlint_source(error,version)
    elif platform_system.startswith("Win"):
        try:
            subprocess.run(["choco", "install", "actionlint", "-y", f"--version='{version}'"], check=True)
            return True, ""
        except (FileNotFoundError, subprocess.CalledProcessError):
            return False, f"{error} : check Choco installation"
    return False, error

def install_actionlint_source(error, version) -> Tuple[bool, str]:
    version = version
    """Install Actionlint Binary from provided script"""
    url = "https://raw.githubusercontent.com/rhysd/actionlint/main/scripts/download-actionlint.bash"
    request = urllib.request.urlopen(url)
    with open("download-actionlint.bash", "wb+") as fp:
        fp.write(request.read())
    try:
        subprocess.run(["bash", "download-actionlint.bash", version], check=True)
        return True, "./actionlint"
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False, error


def check_actionlint_path(platform_system: str, version: str) -> Tuple[bool, str]:
    """Check if the actionlint is in the system's PATH."""
    try:
        installed = subprocess.run(
            ["actionlint", "--version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        if version in f"{installed}":
            return True, ""
        else:
            return install_actionlint(platform_system, version)
    except subprocess.CalledProcessError:
        return (
            False,
            "Failed to install Actionlint, \
please check your package installer or manually install it",
        )
    except FileNotFoundError:
        return check_actionlint_local(platform_system, version)

def check_actionlint_local(platform_system: str, version: str) -> Tuple[bool, str]:

        try:
            installed = subprocess.run(
            ["./actionlint", "--version"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            )
            if version in f"{installed}":
                return True, "./actionlint"
            else:
                return install_actionlint(platform_system, version)
        except FileNotFoundError:
                return install_actionlint(platform_system, version)


class RunActionlint(Rule):
    """Rule to run actionlint as part of workflow linter V2."""

    def __init__(self, settings: Optional[Settings] = None, lint_level: Optional[LintLevels] = LintLevels.NONE) -> None:
        self.message = "Actionlint must pass without errors"
        self.on_fail = lint_level
        self.compatibility = [Workflow]
        self.settings = settings

    def fn(self, obj: Workflow) -> Tuple[bool, str]:
        if not obj or not obj.filename:
            raise AttributeError(
                "Running actionlint without a filename is not currently supported"
            )

        if not self.settings.actionlint_version:
            raise KeyError("The 'actionlint_version' is missing in the configuration file.")

        """Check if Actionlint is alerady installed and if it is installed somewhere not on the PATH (location)"""
        installed, location = check_actionlint_path(platform.system(), self.settings.actionlint_version)
        if installed:
            if location:
                result = subprocess.run(
                     [location, obj.filename],
                    capture_output=True,
                    text=True,
                    check=False,
                )
            else:
                result = subprocess.run(
                    ["actionlint", obj.filename],
                    capture_output=True,
                    text=True,
                    check=False,
                )
            if result.returncode == 1:
                return False, result.stdout
            if result.returncode > 1:
                return False, result.stdout
            return True, ""
        else:
            return False, self.message



"""Module to load for Workflows and Rules."""

import importlib

from typing import List, Optional

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from .models.job import Job
from .models.step import Step
from .models.workflow import Workflow
from .rule import Rule
from .utils import Settings, LintLevels

yaml = YAML()


class WorkflowBuilderError(Exception):
    """Exception to indicate an error with the WorkflowBuilder."""

    pass


class WorkflowBuilder:
    """Collection of methods to build Workflow objects."""

    @classmethod
    def __load_workflow_from_file(cls, filename: str) -> CommentedMap:
        """Load YAML from disk.

        Args:
          filename:
            The name of the YAML file to read.

        Returns:
          A CommentedMap that contains the dict() representation of the
          YAML file. It includes the comments as a part of their respective
          objects (depending on their location in the file).
        """
        with open(filename, encoding="utf8") as file:
            if not file:
                raise WorkflowBuilderError(f"Could not load {filename}")
            try:
                return yaml.load(file)
            except Exception as e:
                raise WorkflowBuilderError(f"Error loading YAML file {filename}: {e}")

    @classmethod
    def __build_workflow(cls, filename: str, loaded_yaml: CommentedMap) -> Workflow:
        """Parse the YAML and build out the workflow to run Rules against.

        Args:
          filename:
            The name of the file that the YAML was loaded from
          loaded_yaml:
            YAML that was loaded from either code or a file

        Returns
          A Workflow to run linting Rules against
        """
        try:
            if not loaded_yaml:
                raise WorkflowBuilderError("No YAML loaded")
            return Workflow.init("", filename, loaded_yaml)
        except Exception as e:
            raise WorkflowBuilderError(f"Error building workflow: {e}")

    @classmethod
    def build(
        cls,
        filename: Optional[str] = None,
        workflow: Optional[CommentedMap] = None,
        from_file: bool = True,
    ) -> Workflow:
        """Build a Workflow from either code or a file.

        This is a method that assists in testing by abstracting the disk IO
        and allows for passing in a YAML object in code.

        Args:
          filename:
            The name of the file to load the YAML workflow from
          yaml:
            Pre-loaded YAML of a workflow
          from_file:
            Flag to determine if the YAML has already been loaded or needs to
            be loaded from disk
        """
        if from_file and filename is not None:
            return cls.__build_workflow(
                filename, cls.__load_workflow_from_file(filename)
            )
        elif not from_file and workflow is not None:
            return cls.__build_workflow("", workflow)

        raise WorkflowBuilderError(
            "The workflow must either be built from a file or from a CommentedMap"
        )


class LoadRulesError(Exception):
    """Exception to indicate an error with loading rules."""

    pass


class Rules:
    """A collection of all of the types of rules.

    Rules is used as a collection of which Rules apply to which parts of the
    workflow. It also assists in making sure the Rules that apply to multiple
    types are not skipped.
    """

    workflow: List[Rule] = []
    job: List[Rule] = []
    step: List[Rule] = []

    def __init__(self, settings: Settings) -> None:
        """Initializes the Rules

        Args:
          settings:
            A Settings object that contains any default, overridden, or custom settings
            required anywhere in the application.
        """
        # [TODO]: data resiliency
        for rule in settings.enabled_rules:
            rule_id = rule["id"]
            module_name = rule_id.split(".")
            module_name = ".".join(module_name[:-1])
            rule_name = rule_id.split(".")[-1]

            try:
                rule_class = getattr(importlib.import_module(module_name), rule_name)
                rule_inst = rule_class(settings=settings, lint_level=lint_level(rule["level"]))

                if Workflow in rule_inst.compatibility:
                    self.workflow.append(rule_inst)
                if Job in rule_inst.compatibility:
                    self.job.append(rule_inst)
                if Step in rule_inst.compatibility:
                    self.step.append(rule_inst)
            except LoadRulesError as err:
                print(f"Error loading: {rule}\n{err}")

    def list(self) -> None:
        """Print the loaded Rules."""
        print("===== Loaded Rules =====")
        print("workflow rules:")
        for rule in self.workflow:
            print(f" - {type(rule).__name__}")
        print("job rules:")
        for rule in self.job:
            print(f" - {type(rule).__name__}")
        print("step rules:")
        for rule in self.step:
            print(f" - {type(rule).__name__}")
        print("========================\n")


def lint_level(level: str) -> LintLevels:
    """Convert a string to a LintLevels enum.

    Args:
      level:
        The string representation of the LintLevels enum

    Returns:
      The LintLevels enum value
    """
    if level == "error":
        return LintLevels.ERROR
    elif level == "warning":
        return LintLevels.WARNING
    else:
        return LintLevels.NONE



"""A Rule to enforce Actions are pinned correctly."""

from typing import Optional, Tuple

from ..models.step import Step
from ..rule import Rule
from ..utils import LintLevels, Settings


class RuleStepUsesPinned(Rule):
    """Rule to contain the enforcement logic for pinning Actions versions.

    Definition of Internal Action:
      An Action that exists in the `bitwarden/gh-actions` GitHub Repository.

    For any external Action (any Action that does not fit the above definition of
    an Internal Action), to mitigate the risks of supply chain attacks in our CI
    pipelines, we pin any use of an Action to a specific hash that has been verified
    and pre-approved after a security audit of the version of the Action.

    All Internal Actions, should be pinned to 'main'. This prevents Renovate from
    spamming a bunch of PRs across all of our repos when `bitwarden/gh-actions` is
    updated.
    """

    def __init__(self, settings: Optional[Settings] = None, lint_level: Optional[LintLevels] = LintLevels.ERROR) -> None:
        """Constructor for RuleStepUsesPinned to override base Rule.

        Args:
          settings:
            A Settings object that contains any default, overridden, or custom settings
            required anywhere in the application.
          lint_level:
            The LintLevels enum value to determine the severity of the rule.
        """
        self.on_fail = lint_level
        self.compatibility = [Step]
        self.settings = settings

    def skip(self, obj: Step) -> bool:
        """Skip this Rule on some Steps.

        This Rule does not apply to a few types of Steps. These
        Rules are skipped.
        """
        if not obj.uses:
            return True

        ## Force pass for any local actions
        if "@" not in obj.uses:
            return True

        return False

    def fn(self, obj: Step) -> Tuple[bool, str]:
        """Enforces all Actions to be pinned in a specific way.

        Pinning external Action hashes prevents unknown updates that could
        break the pipelines or be the entry point to a supply chain attack.

        Pinning internal Actions to branches allow for less updates as work
        is done on those repos. This is mainly to support our Action
        monorepo architecture of our Actions.

        Example:
        - name: Checkout Branch
          uses: actions/checkout@b4ffde65f46336ab88eb53be808477a3936bae11 # v4.1.1

        - name: Test Bitwarden Action
          uses: bitwarden/gh-actions/get-keyvault-secrets@main

        - name: Test Local Action
          uses: ./actions/test-action

        - name: Test Run Action
          run: echo "test"

        In this example, 'actions/checkout' must be pinned to the full commit
        of the tag while 'bitwarden/gh-actions/get-keyvault-secrets' must be
        pinned to 'main'. The other two Steps will be skipped.
        """
        if self.skip(obj):
            return True, ""

        path, ref = obj.uses.split("@")

        if path.startswith("bitwarden/"):
            if ref == "main" or "sm-action" in path:
                return True, ""
            return False, "Please pin to main"

        try:
            int(ref, 16)
        except ValueError:
            return False, "Please pin the action to a commit sha"

        if len(ref) != 40:
            return False, "Please use the full commit sha to pin the action"

        if not obj.uses_comment:
            return False, "Please comment the version of the action commit sha"

        return True, ""



"""Module providing Lint subcommand to run custom linting rules against GitHub Action
Workflows."""

import argparse
import os

from typing import Optional

from .load import WorkflowBuilder, Rules
from .utils import LintFinding, Settings


class LinterCmd:
    """Command to lint GitHub Action Workflow files

    This class contains logic to lint workflows that are passed in.
    Supporting logic is supplied to:
      - build out the list of Rules desired
      - select and validate the workflow files to lint
    """

    def __init__(self, settings: Optional[Settings] = None) -> None:
        """Initailize the LinterCmd class.

        Args:
          settings:
            A Settings object that contains any default, overridden, or custom settings
            required anywhere in the application.
        """
        self.rules = Rules(settings=settings)

    @staticmethod
    def extend_parser(
        subparsers: argparse._SubParsersAction,
    ) -> argparse._SubParsersAction:
        """Extends the CLI subparser with the options for LintCmd.

        Add 'lint' as a subcommand along with its options and arguments

        Args:
          subparsers:
            The main argument parser to add subcommands and arguments to
        """
        parser_lint = subparsers.add_parser(
            "lint",
            help="Verify that a GitHub Action Workflow follows all of the Rules.",
        )
        parser_lint.add_argument(
            "-s",
            "--strict",
            action="store_true",
            help="return non-zero exit code on warnings as well as errors",
        )
        parser_lint.add_argument("-f", "--files", nargs="+", action="append", required=True, help="files to lint")
        parser_lint.add_argument(
            "-o",
            "--output",
            action="store",
            help="output format: [stdout|json|md]",
            default="stdout",
        )
        return subparsers

    def get_max_error_level(self, findings: list[LintFinding]) -> int:
        """Get max error level from list of findings.

        Compute the maximum error level to determine the exit code required.
        # if max(error) return exit(1); else return exit(0)

        Args:
          findings:
            All of the findings that the linter found while linting a workflow.

        Return:
          The numeric value of the maximum lint finding
        """
        if len(findings) == 0:
            return 0
        return max(findings, key=lambda finding: finding.level.code).level.code

    def lint_file(self, filename: str) -> int:
        """Lint a single workflow.

        Run all of the Workflow, Job, and Step level rules that have been enabled.

        Args:
          filename:
            The name of the file that contains the workflow to lint

        Returns:
          The maximum error level found in the file (none, warning, error) to
          calculate the exit code from.
        """
        findings = []
        max_error_level = 0

        print(f"Linting: {filename}")
        workflow = WorkflowBuilder.build(filename)

        for rule in self.rules.workflow:
            findings.append(rule.execute(workflow))

        for _, job in workflow.jobs.items():
            for rule in self.rules.job:
                findings.append(rule.execute(job))

            if job.steps is not None:
                for step in job.steps:
                    for rule in self.rules.step:
                        findings.append(rule.execute(step))

        findings = list(filter(lambda a: a is not None, findings))

        if len(findings) > 0:
            for finding in findings:
                print(f" - {finding}")
            print(f"Issues found by {len(findings)} rules in {filename}")
            print()

        max_error_level = self.get_max_error_level(findings)

        return max_error_level

    def generate_files(self, files: list[str]) -> list[str]:
        """Generate the list of files to lint.

        Searches the list of directory and/or files taken from the CLI.

        Args:
          files:
            list of file names or directory names.

        Returns:
          A sorted set of all workflow files in the path(s) specified.
        """
        workflow_files = []
        for path in files:
            if os.path.isfile(path):
                workflow_files.append(path)
            elif os.path.isdir(path):
                for dirpath, dirnames, filenames in os.walk(path):
                    dirnames[:] = []
                    for file in filenames:
                        filepath = dirpath + os.sep + file
                        if filepath.endswith((".yml", ".yaml")):
                            workflow_files.append(filepath)

        return sorted(set(workflow_files))

    def run(self, input_files: list[str], strict: bool = False) -> int:
        """Execute the LinterCmd.

        Args:
          input_files:
            list of file names or directory names.
          strict:
            fail on WARNING instead of succeed

        Returns
          The return_code for the entire CLI to indicate success/failure
        """
        files = self.generate_files(input_files)

        if len(input_files) > 0:
            files_with_issues = []
            return_code = 0
            for file in files:
                return_value = self.lint_file(file)
                if return_value > 0:
                    files_with_issues.append(file)
                    return_code = max(return_code, return_value)

            if len(files_with_issues) > 0:
                newline = "\n"  # For compatibility with Python 3.11
                print(
                    f"""Found {len(files_with_issues)} file(s) with issues:
  {f"{newline}  ".join(files_with_issues)}

For help, refer to
  - Workflow Syntax: \
https://docs.github.com/en/actions/writing-workflows/workflow-syntax-for-github-actions
  - Bitwarden Examples: \
https://github.com/bitwarden/workflow-linter/tree/main/.github/workflows/examples")"""
                )
            else:
                print("No issues found")

            if return_code == 1 and not strict:
                return_code = 0

            return return_code
        else:
            print(f'File(s)/Directory: "{input_files}" does not exist, exiting.')
            return -1



"""Tests src/bitwarden_workflow_linter/load.py."""

import pytest

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from .conftest import FIXTURE_DIR

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.models.workflow import Workflow


yaml = YAML()


@pytest.fixture(name="workflow_filename")
def fixture_workflow_filename():
    return f"{FIXTURE_DIR}/test.yml"


@pytest.fixture(name="simple_workflow_yaml")
def fixture_simple_workflow_yaml():
    return yaml.load(
        """\
---
name: test
on:
  workflow_dispatch:

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - name: Test
        run: echo test
"""
    )


@pytest.fixture(name="complex_workflow_yaml")
def fixture_complex_workflow_yaml():
    return yaml.load(
        """\
---
name: test
on:
  workflow_dispatch:

jobs:
  job-key:
    name: Test
    runs-on: ubuntu-latest
    steps:
      - name: Test
        run: echo test

  call-workflow:
    uses: bitwarden/server/.github/workflows/workflow-linter.yml@master

  test-normal-action:
    name: Download Latest
    runs-on: ubuntu-20.04
    steps:
      - name: Checkout
        uses: actions/checkout@2541b1294d2704b0964813337f33b291d3f8596b

      - run: |
          echo test

  test-local-action:
    name: Testing a local action call
    runs-on: ubuntu-20.04
    steps:
      - name: local-action
        uses: ./version-bump
"""
    )


def test_load_workflow_from_file(workflow_filename: str) -> None:
    workflow = WorkflowBuilder.build(workflow_filename)
    assert isinstance(workflow, Workflow)


def test_load_simple_workflow_from_yaml(simple_workflow_yaml: CommentedMap) -> None:
    workflow = WorkflowBuilder.build(workflow=simple_workflow_yaml, from_file=False)
    assert isinstance(workflow, Workflow)


def test_load_complex_workflow_from_yaml(complex_workflow_yaml: CommentedMap) -> None:
    workflow = WorkflowBuilder.build(workflow=complex_workflow_yaml, from_file=False)
    assert isinstance(workflow, Workflow)



"""Test src/bitwarden_workflow_linter/rules/pinned_job_runner.py."""

import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.load import WorkflowBuilder
from src.bitwarden_workflow_linter.rules.pinned_job_runner import (
    RuleJobRunnerVersionPinned,
)

yaml = YAML()


@pytest.fixture(name="correct_runner")
def fixture_correct_runner():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-22.04
    steps:
      - run: echo test

  call-workflow:
    uses: bitwarden/server/.github/workflows/workflow-linter.yml@master
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="incorrect_runner")
def fixture_incorrect_runner():
    workflow = """\
---
on:
  workflow_dispatch:

jobs:
  job-key:
    runs-on: ubuntu-latest
    steps:
      - run: echo test
"""
    return WorkflowBuilder.build(workflow=yaml.load(workflow), from_file=False)


@pytest.fixture(name="rule")
def fixture_rule():
    return RuleJobRunnerVersionPinned()


def test_rule_on_correct_runner(rule, correct_runner):
    result, _ = rule.fn(correct_runner.jobs["job-key"])
    assert result is True

    result, _ = rule.fn(correct_runner.jobs["call-workflow"])
    assert result is True


def test_rule_on_incorrect_runner(rule, incorrect_runner):
    result, _ = rule.fn(incorrect_runner.jobs["job-key"])
    assert result is False



"""Test src/bitwarden_workflow_linter/models/step.py."""

import json
import pytest

from ruamel.yaml import YAML

from src.bitwarden_workflow_linter.models.step import Step


@pytest.fixture(name="default_step")
def fixture_default_step():
    step_str = """\
name: Default Step
run: echo "test"
"""
    yaml = YAML()
    step_yaml = yaml.load(step_str)
    return Step.init(0, "default", step_yaml)


@pytest.fixture(name="uses_step")
def fixture_uses_step():
    step_str = """\
name: Download Artifacts
uses: bitwarden/download-artifacts@main # v1.0.0
with:
    workflow: upload-test-artifacts.yml
    artifacts: artifact
    path: artifact
    branch: main

"""
    yaml = YAML()
    step_yaml = yaml.load(step_str)
    return Step.init(0, "default", step_yaml)


@pytest.fixture(name="uses_step_no_comments")
def fixture_uses_step_no_comments():
    step_str = """\
name: Download Artifacts
uses: bitwarden/download-artifacts@main
with:
    workflow: upload-test-artifacts.yml
    artifacts: artifact
    path: artifact
    branch: main

"""
    yaml = YAML()
    step_yaml = yaml.load(step_str)
    return Step.init(0, "default", step_yaml)


@pytest.fixture(name="uses_step_no_ref")
def fixture_uses_step_no_ref():
    step_str = """\
name: Run Local Workflow
uses: ./.github/workflows/_local.yml
"""
    yaml = YAML()
    step_yaml = yaml.load(step_str)
    return Step.init(0, "default", step_yaml)


@pytest.fixture(name="uses_step_no_ref_with_comments")
def fixture_uses_step_no_ref_with_comments():
    step_str = """\
name: Run Local Workflow
uses: ./.github/workflows/_local.yml # A comment
"""
    yaml = YAML()
    step_yaml = yaml.load(step_str)
    return Step.init(0, "default", step_yaml)


def test_step_default(default_step):
    assert default_step.key == 0
    assert default_step.job == "default"
    assert default_step.name == "Default Step"
    assert default_step.env is None
    assert default_step.uses is None
    assert default_step.uses_with is None
    assert default_step.run == 'echo "test"'


def test_step_no_keyword_field(default_step):
    assert default_step.uses_with is None
    assert "uses_with" not in default_step.to_json()


def test_step_extra_kwargs(default_step):
    with pytest.raises(Exception):
        assert default_step.extra == "test"


def test_step_keyword_field(uses_step):
    expected_response = {
        "workflow": "upload-test-artifacts.yml",
        "artifacts": "artifact",
        "path": "artifact",
        "branch": "main",
    }

    step_json = uses_step.to_json()
    assert uses_step.key == 0
    assert "uses_with" not in step_json
    assert "with" in step_json
    assert json.loads(uses_step.to_json())["with"] == expected_response


def test_step_comment(uses_step):
    assert uses_step.key == 0
    assert uses_step.job == "default"
    assert uses_step.uses_ref == "main"
    assert uses_step.uses_version == "v1.0.0"
    assert uses_step.uses_comment == "# v1.0.0"


def test_step_no_comments(uses_step_no_comments):
    assert uses_step_no_comments.key == 0
    assert uses_step_no_comments.job == "default"
    assert uses_step_no_comments.uses_ref == "main"
    assert uses_step_no_comments.uses_version is None
    assert uses_step_no_comments.uses_comment is None


def test_step_no_ref(uses_step_no_ref):
    assert uses_step_no_ref.key == 0
    assert uses_step_no_ref.job == "default"
    assert uses_step_no_ref.uses_ref is None
    assert uses_step_no_ref.uses_version is None
    assert uses_step_no_ref.uses_comment is None


def test_step_no_ref_with_comments(uses_step_no_ref_with_comments):
    assert uses_step_no_ref_with_comments.key == 0
    assert uses_step_no_ref_with_comments.job == "default"
    assert uses_step_no_ref_with_comments.uses_ref is None
    assert (
        uses_step_no_ref_with_comments.uses_version == "comment"
    )  # We are not currently validating the version matches a specific format
    assert uses_step_no_ref_with_comments.uses_comment == "# A comment"
